package checkers.interned;

import static java.util.Collections.singleton;

import javax.lang.model.element.*;

import checkers.basetype.BaseTypeChecker;
import checkers.flow.Flow;
import checkers.interned.quals.Interned;
import checkers.types.*;
import static checkers.types.AnnotatedTypeMirror.*;

import com.sun.source.tree.*;

/**
 * An {@link AnnotatedTypeFactory} that accounts for the properties of the
 * Interned type system. This type factory will add the {@link Interned}
 * annotation to a type if the input:
 *
 * 1. is a String literal
 * 2. is a class literal
 * 3. has an enum type
 * 4. has a primitive type
 * 5. has the type java.lang.Class
 * 6. is a call to the method {@link String#intern()}
 */
public class InternedAnnotatedTypeFactory extends AnnotatedTypeFactory {

    /** Adds annotations from tree context before type resolution. */
    private final TreeAnnotator treeAnnotator;

    /** Adds annotations from the resulting type after type resolution. */
    private final TypeAnnotator typeAnnotator;

    /** The {@link Interned} annotation. */
    final AnnotationMirror INTERNED;

    /** Flow-sensitive qualifier inference. */
    private final Flow flow;

    /** An element of the class {@code java.lang.Class} */
    private final TypeElement elementOfClass;

    /** An element of class {@code java.lang.String} */
    private final TypeElement elementOfString;
    /**
     * Creates a new {@link InternedAnnotatedTypeFactory} that operates on a
     * particular AST.
     *
     * @param checker the checker to use
     * @param root the AST on which this type factory operates
     */
    public InternedAnnotatedTypeFactory(InternedChecker checker,
        CompilationUnitTree root) {
        super(checker, root);
        this.INTERNED = annotations.fromClass(Interned.class);
        this.elementOfClass = elements.getTypeElement("java.lang.Class");
        this.elementOfString = elements.getTypeElement("java.lang.String");

        this.treeAnnotator = new TreeAnnotator(checker);
        this.typeAnnotator = new InternedTypeAnnotator(checker);

        this.flow = new Flow(checker, root, singleton(INTERNED), this);
        if (checker.getLintOption("flow", true)) flow.scan(root, null);
    }

    @Override
    protected void annotateImplicit(Element elt, AnnotatedTypeMirror type) {
        typeAnnotator.visit(type);
    }

    @Override
    protected void annotateImplicit(Tree tree, AnnotatedTypeMirror type) {
        treeAnnotator.visit(tree, type);
        final AnnotationMirror result = flow.test(tree);
        if (result != null)
            type.addAnnotation(result);
        typeAnnotator.visit(type);
    }

    /**
     * A class for adding annotations to a type after initial type resolution.
     */
    private class InternedTypeAnnotator extends TypeAnnotator {

        /** Creates an {@link InternedTypeAnnotator} for the given checker. */
        InternedTypeAnnotator(BaseTypeChecker checker) {
            super(checker);
        }

        @Override
        public Void visitDeclared(AnnotatedDeclaredType t, Void p) {

            // Enum types and class: add an @Interned annotation.
            Element elt = t.getUnderlyingType().asElement();
            assert elt != null;
            if ((elt.getKind() == ElementKind.ENUM)
                    || (elementOfClass.equals(elt)))
                t.addAnnotation(INTERNED);

            return super.visitDeclared(t, p);
        }

        @Override
        public Void visitExecutable(AnnotatedExecutableType t, Void p) {
            // Annotate the java.lang.String.intern() method.
            ExecutableElement method = t.getElement();
            if (method.getSimpleName().contentEquals("intern")
                    && elementOfString.equals(method.getEnclosingElement()))
                t.getReturnType().addAnnotation(INTERNED);

            return super.visitExecutable(t, p);
        }
    }

    @Override
    public AnnotatedPrimitiveType getUnboxedType(AnnotatedDeclaredType type) {
        AnnotatedPrimitiveType primitive = super.getUnboxedType(type);
        primitive.addAnnotation(INTERNED);
        return primitive;
    }
}
