package checkers.javari;

import java.util.*;

import javax.lang.model.element.*;

import com.sun.source.tree.*;
import com.sun.source.util.SimpleTreeVisitor;
import com.sun.tools.javac.tree.JCTree;

import checkers.javari.quals.*;
import checkers.util.*;
import checkers.types.*;

import static checkers.types.AnnotatedTypeMirror.*;


/**
 * Generates a AnnotatedTypeMirror with Javari annotations from a Tree
 * or a Element parameter.
 * <p>
 * Implicit annotations are added as follows:
 * <ul>
 *
 *   <li> Qualified class types without annotations receive the
 *   {@code @Mutable} annotation.
 *
 *   <li> Qualified executable types receivers without annotations
 *   are annotated with the qualified executable type owner's
 *   annotation.
 *
 *   <li> Qualified declared types are annotated with their
 *   underlying type's element annotations.
 *
 *   <li> Qualified types whose elements correspond to fields, and
 *   all its subtypes, are annotated with {@code @ReadOnly},
 *   {@code @Mutable} or {@code @PolyRead), according to the
 *   qualified type of {@code this}.
 *
 *</ul>
 */
public class JavariAnnotatedTypeFactory extends AnnotatedTypeFactory {

    /** Adds annotations from tree context before type resolution. */
    private final JavariTreePreAnnotator treePre;

    /** Adds annotations from the resulting type after type resolution. */
    private final JavariTypePostAnnotator typePost;

    /** The Javari annotations. */
    private final AnnotationMirror READONLY, THISMUTABLE, MUTABLE, POLYREAD, QREADONLY, ASSIGNABLE;

    /**
     * Creates a new {@link JavariAnnotatedTypeFactory} that operates on a
     * particular AST.
     *
     * @param checker the checker to which this factory belongs
     * @param root the AST on which this type factory operates
     */
    public JavariAnnotatedTypeFactory(JavariChecker checker,
        CompilationUnitTree root) {
        super(checker, root);
        this.treePre = new JavariTreePreAnnotator();
        this.typePost = new JavariTypePostAnnotator();
        this.READONLY = checker.READONLY;
        this.THISMUTABLE = checker.THISMUTABLE;
        this.MUTABLE = checker.MUTABLE;
        this.POLYREAD = checker.POLYREAD;
        this.QREADONLY = checker.QREADONLY;
        this.ASSIGNABLE = checker.ASSIGNABLE;
    }

    /**
     * Returns the annotation specifying the immutability type of {@code type}.
     */
    private AnnotationMirror getImmutabilityAnnotation(/*@ReadOnly*/ AnnotatedTypeMirror type) {
        Set<AnnotationMirror> annotations = new HashSet<AnnotationMirror>(type.getAnnotations());
        annotations.remove(ASSIGNABLE);
    	return qualHierarchy.validQualifier(annotations);
    }

    /**
     * @param type  an annotated type mirror
     * @return true iff the type is specified an immutability type
     * other than this-mutable, false otherwise
     */
    public boolean hasImmutabilityAnnotation(/*@ReadOnly*/ AnnotatedTypeMirror type) {
        return type != null && getImmutabilityAnnotation(type) != null;
    }


    /**
     * Adds implicit annotations to a qualified type, based on its
     * tree, as follows:
     * 
     * <ul>
     *
     *   <li> Resolves qualified types from MemberSelectTree,
     *    inheritting from the expression to the identifier if the
     *    identifier is {@code @ThisMutable}.
     *
     *   <li> Qualified class types without annotations receive the
     *    {@code @Mutable} annotation.
     *
     *   <li> Qualified executable types receivers without annotations
     *    are annotated with the qualified executable type owner's
     *    annotation.
     *
     *   <li> Qualified declared types are annotated with their
     *    underlying type's element annotations.
     *
     *   <li> Qualified types whose elements correspond to fields, and
     *    all its subtypes, are annotated with {@code @ReadOnly},
     *    {@code @Mutable} or {@code @PolyRead), according to the
     *    qualified type of {@code this}.
     *
     * </ul>
     *
     * @param tree an AST node
     * @param type the type obtained from {@code tree}
     */
    @Override
    protected void annotateImplicit(Tree tree, /*\@Mutable*/ AnnotatedTypeMirror type) {
        treePre.visit(tree, type);
        if (tree.getKind() == Tree.Kind.CLASS
            && !hasImmutabilityAnnotation(type))
            type.addAnnotation(MUTABLE);
        else
            typePost.visit(type);

        if (type.hasAnnotation(THISMUTABLE)) {

            AnnotatedDeclaredType selfType = getSelfType(tree);
            if (selfType != null) {
                if (selfType.hasAnnotation(POLYREAD)) {
                    new SimpleAnnotatedTypeScanner<Void, Void>() {
                        @Override
                            protected Void defaultAction(AnnotatedTypeMirror type, Void p) {
                            if (type.hasAnnotation(THISMUTABLE)) {
                                type.removeAnnotation(THISMUTABLE);
                                type.addAnnotation(POLYREAD);
                            }
                            return super.defaultAction(type, p);
                        }
                    }.visit(type);
                }
                else if (selfType.hasAnnotation(MUTABLE)) {
                    new SimpleAnnotatedTypeScanner<Void, Void>() {
                        @Override
                            protected Void defaultAction(AnnotatedTypeMirror type, Void p) {
                            if (type.hasAnnotation(THISMUTABLE)) {
                                type.removeAnnotation(THISMUTABLE);
                                type.addAnnotation(MUTABLE);
                            }
                            return super.defaultAction(type, p);
                        }
                    }.visit(type);
                }
            }
        }
    }

    /**
     * Adds annotations to qualified types according to their provided
     * element, as follows:
     *
     * <ul>
     *
     *   <li> Qualified class types without annotations corresponding
     *   to class or interface elements receive the {@code @Mutable}
     *   annotation.
     *
     *   <li> Unannotated receivers of qualified executable types are
     *   annotated with the qualified type owner's annotation.
     *
     *   <li> Unannotated qualified declared types are annotated with
     *   their underlying type's element annotations.
     *
     *   <li> Qualified types whose elements correspond to fields, and
     *   all its subtypes, are annotated with {@code @ReadOnly} or
     *   {@code @ThisMutable}, according to the supertype.
     *
     * </ul>
     *
     * @param element an element
     * @param type the type obtained from {@code elt}
     */
    @Override
    protected void annotateImplicit(Element element, /*\@Mutable*/ AnnotatedTypeMirror type) {
        if (element.getKind().isClass() || element.getKind().isInterface()) {
            if (!hasImmutabilityAnnotation(type))
                type.addAnnotation(MUTABLE);
        } else
            typePost.visit(type);

    }

    /**
     * Does nothing.
     *
     * Don't use the framework to inherit annotations from classes;
     * Javari behavior on annotation inheritance is ad hoc enough,
     * so we instead implement it all on this class, and override this
     * method with an empty method to remove the framework behavior.
     */
    @Override
    protected void annotateInheritedFromClass(/*\@Mutable*/ AnnotatedTypeMirror type) {
        // Do nothing
    }

    /**
     * Returns a singleton collection with the most restrictive immutability
     * annotation that is a supertype of the annotations on both collections.
     */
    @Override
    public Collection<AnnotationMirror> unify(Collection<AnnotationMirror> c1,
            Collection<AnnotationMirror> c2) {
        Map<String, AnnotationMirror> ann =
            new HashMap<String, AnnotationMirror>();
        for (AnnotationMirror anno : c1)
            ann.put(AnnotationUtils.annotationName(anno).toString(), anno);
        for (AnnotationMirror anno : c2)
            ann.put(AnnotationUtils.annotationName(anno).toString(), anno);

        if (ann.containsKey(QReadOnly.class.getCanonicalName()))
            return Collections.singleton(QREADONLY);
        else if (ann.containsKey(ReadOnly.class.getCanonicalName()))
            return Collections.singleton(READONLY);
        else if (ann.containsKey(PolyRead.class.getCanonicalName()))
            return Collections.singleton(POLYREAD);
        else
            return Collections.singleton(MUTABLE);
    }

    /**
     * Determines the type of the constructed object based on the
     * parameters passed to the constructor. The new object has the
     * same mutability as the annotation marked on the constructor
     * receiver, as resolved by this method.
     *
     * {@code @PolyRead} receiver values are resolved by looking at
     * the mutability of any parameters marked as {@code
     * @PolyRead}. The rules are similar to the ones applicable to
     * method invocation resolution, but without looking at {@code
     * @this}.
     *
     * <ul>
     *
     *  <li> If all parameters marked as {@code @PolyRead} receive
     *  {@code @Mutable} arguments, the receiver value is resolved as
     *  {@code @Mutable}.
     *
     *  <li> If all parameters marked as {@code @PolyRead} receive
     *  {@code @Mutable} or {@code @ThisMutable} arguments and the
     *  condition above is not satisfied, the receiver value is
     *  resolved as {@code @ThisMutable}.
     *
     *  <li> If all parameters marked as {@code @PolyRead} receive
     *  {@code @Mutable} or {@code @PolyRead} arguments and none of
     *  the condition above is satisfied, the receiver value is
     *  resolved as {@code @PolyRead}.
     *
     *  <li> If none of the conditions above is satisfied, the
     *  receiver value is resolved as {@code @ReadOnly}.
     *
     * </ul>
     *
     * @param tree  the new class tree
     * @return AnnotatedExecutableType corresponding to the type being
     * constructed, with the resolved type on its receiver.
     */
    @Override
    public AnnotatedExecutableType constructorFromUse(NewClassTree tree) {

        ExecutableElement executableElt = InternalUtils.constructor(tree);
        AnnotatedExecutableType exType = fromElement(executableElt);
        AnnotatedTypeMirror receiverType = exType.getReceiverType();
        // use the original type annotations
        annotateImplicit(tree.getIdentifier(), receiverType);

        List<AnnotatedTypeMirror> argumentTypes = atypes.getAnnotatedTypes(tree.getArguments());
        List<AnnotatedTypeMirror> parameterTypes = atypes.expandVarArgs(exType, tree.getArguments());

        boolean allMutable = true, allPolyRead = true, allThisMutable = true;

        // look at parameters and arguments
        for (int i = 0; i < parameterTypes.size(); i++) {
            AnnotatedTypeMirror pType = parameterTypes.get(i);
            if (pType.hasAnnotation(POLYREAD)) {
                AnnotatedTypeMirror aType = argumentTypes.get(i);

                if (aType.hasAnnotation(THISMUTABLE) || aType.hasAnnotation(POLYREAD))
                    allMutable = false;

                if (aType.hasAnnotation(READONLY) || aType.hasAnnotation(QREADONLY)) {
                    allMutable = false; allThisMutable = false;
                }

                if (!(aType.hasAnnotation(POLYREAD)
                      && !aType.hasAnnotation(READONLY)
                      && !aType.hasAnnotation(THISMUTABLE)
                      && !aType.hasAnnotation(QREADONLY)))
                    allPolyRead = false;
            }
        }

        final AnnotationMirror replacement;
        if (allMutable) replacement = MUTABLE;
        else if (allThisMutable) replacement = THISMUTABLE;
        else if (allPolyRead) replacement = POLYREAD;
        else replacement = READONLY;
        
        if (replacement != POLYREAD) {
	        AnnotatedTypeVisitor<Void, Void> replacer = new SimpleAnnotatedTypeScanner<Void, Void>() {
	            @Override
	                protected Void defaultAction(AnnotatedTypeMirror type, Void p) {
	                if (type.hasAnnotation(POLYREAD)) {
	                    type.removeAnnotation(POLYREAD);
	                    type.addAnnotation(replacement);
	                }
	                return super.defaultAction(type, p);
	            }
	        };
	        
	        for (AnnotatedTypeMirror t : exType.getParameterTypes()) replacer.visit(t);
	        replacer.visit(receiverType);
        }

        return exType;
    }

    /**
     * Determines the type of the invoked method based on the passed method
     * invocation tree.
     *
     * Invokes the super method, then resolves annotations {@code
     * @PolyRead} at the raw level on return values by looking at the
     * mutability of any parameters marked as {@code @PolyRead}. For
     * this purpose, a {@code @PolyRead} annotation on the receiver
     * counts as if {@code this} were being passed as an argument to a
     * parameter marked as {@code @PolyRead}.
     *
     * <ul>
     *
     *  <li> If all parameters marked as {@code @PolyRead} receive
     *  {@code @Mutable} arguments, the return value is resolved as
     *  {@code @Mutable}.
     *
     *  <li> If all parameters marked as {@code @PolyRead} receive
     *  {@code @Mutable} or {@code @ThisMutable} arguments and the
     *  condition above is not satisfied, the return value is
     *  resolved as {@code @ThisMutable}.
     *
     *  <li> If all parameters marked as {@code @PolyRead} receive
     *  {@code @Mutable} or {@code @PolyRead} arguments and none of
     *  the condition above is satisfied, the return value is resolved
     *  as {@code @PolyRead}.
     *
     *  <li> If none of the conditions above is satisfied, the return
     *  value is resolved as {@code @ReadOnly}.
     *
     * </ul>
     *
     * @param tree  the method invocation tree
     * @return AnnotatedExecutableType with return value resolved as described.
     */
    @Override
    public AnnotatedExecutableType methodFromUse(MethodInvocationTree tree) {
        AnnotatedExecutableType type = super.methodFromUse(tree);

        ExecutableElement executableElt = type.getElement();
        AnnotatedTypeMirror returnType = type.getReturnType();

        List<AnnotatedTypeMirror> argumentTypes = atypes.getAnnotatedTypes(tree.getArguments());
        List<AnnotatedTypeMirror> parameterTypes = atypes.expandVarArgs(type, tree.getArguments());
        AnnotatedTypeMirror receiverType = type.getReceiverType();

        boolean allMutable = true, allPolyRead = true, allThisMutable = true;

        // look at parameters and arguments
        for (int i = 0; i < parameterTypes.size(); i++) {
            AnnotatedTypeMirror pType = parameterTypes.get(i);
            // only look at it if parameter is PolyRead
            if (pType.hasAnnotation(POLYREAD)) {
                AnnotatedTypeMirror aType = argumentTypes.get(i);

                if (aType.hasAnnotation(THISMUTABLE) || aType.hasAnnotation(POLYREAD))
                    allMutable = false;

                if (aType.hasAnnotation(READONLY) || aType.hasAnnotation(QREADONLY)) {
                    allMutable = false; allThisMutable = false;
                }

                if (!(aType.hasAnnotation(POLYREAD)
                      && !aType.hasAnnotation(READONLY)
                      && !aType.hasAnnotation(THISMUTABLE)
                      && !aType.hasAnnotation(QREADONLY)))
                    allPolyRead = false;

            }
        }

        // look at receiver type and reference
        if (receiverType.hasAnnotation(POLYREAD)) {
            // if MemberSelectTree, we can just look at the expression tree
            ExpressionTree exprTree = tree.getMethodSelect();
            if (exprTree.getKind() == Tree.Kind.MEMBER_SELECT) {
                MemberSelectTree mst = (MemberSelectTree) exprTree;
                AnnotatedTypeMirror referenceType = getAnnotatedType(mst.getExpression());

                if (referenceType.hasAnnotation(READONLY)) {
                    allMutable = false;
                    allThisMutable = false;
                    allPolyRead = false;
                } else if (referenceType.hasAnnotation(POLYREAD)) {
                    allMutable = false;
                }

            } else {
                // not MemberSelectTree, get context from method's enclosing class
                AnnotatedTypeMirror classType
                    = fromElement(executableElt.getEnclosingElement());
                if (classType.hasAnnotation(READONLY)) {
                    allMutable = false;
                    allThisMutable = false;
                    allPolyRead = false;
                } else if (classType.hasAnnotation(POLYREAD)) {
                    allMutable = false;
                    allThisMutable = false;
                }
            }
        }

        final AnnotationMirror replacement;
        if (allMutable) replacement = MUTABLE;
        else if (allThisMutable) replacement = THISMUTABLE;
        else if (allPolyRead) replacement = POLYREAD;
        else replacement = READONLY;
        
        if (!hasImmutabilityAnnotation(returnType) && !returnType.getKind().isPrimitive())
        	returnType.addAnnotation(THISMUTABLE);
        
        if (replacement != POLYREAD) {
	        AnnotatedTypeVisitor<Void, Void> replacer = new SimpleAnnotatedTypeScanner<Void, Void>() {
	            @Override
	                protected Void defaultAction(AnnotatedTypeMirror type, Void p) {
	                if (type.hasAnnotation(POLYREAD)) {
	                    type.removeAnnotation(POLYREAD);
	                    type.addAnnotation(replacement);
	                }
	                return super.defaultAction(type, p);
	            }
	        };
	        
	        replacer.visit(returnType);
	        for (AnnotatedTypeMirror t : type.getParameterTypes()) replacer.visit(t);
	        replacer.visit(receiverType);
        }
        return type;
    }

    /**
     * We modify this callback method to replace {@code @ThisMutable}
     * implicit annotations with the qualified supertype annotation,
     * if the owner doesn't have a {@code @ReadOnly} annotation.
     * <p>
     * Note on the given example that, if {@code @ThisMutable tmObject} 
     * were resolved as {@code @ReadOnly tmObject}, the code snippet 
     * would be legal. Such a class could then be created to obtain {@code
     * @Mutable} access to {@code tmObject} from a {@code @ReadOnly}
     * reference to it, without typechecker errors.
     * <p>
     * <code>
     *  @PolyRead Object breakJavari(@PolyRead Object s) @ReadOnly {
     *   tmObject = s;
     *   return null;
     *  }
     * </code>
     *
     * @param type  the annotated type of the element
     * @param owner the annotated type of the receiver of the accessing tree
     * @param element   the element of the field or method
     */
    @Override
    public void postAsMemberOf(AnnotatedTypeMirror type,
            AnnotatedTypeMirror owner, Element element) {
        if (!owner.hasAnnotation(READONLY)) {
            final AnnotationMirror ownerAnno = getImmutabilityAnnotation(owner);
            new SimpleAnnotatedTypeScanner<Void, Void>() {
                @Override
                protected Void defaultAction(AnnotatedTypeMirror type, Void p) {
                    if (type.hasAnnotation(THISMUTABLE)) {
                        type.removeAnnotation(THISMUTABLE);
                        type.addAnnotation(ownerAnno);
                    }
                    return super.defaultAction(type, p);
                }
            }.visit(type);
        }
    }


    /**
     * A visitor to get annotations from a tree.
     *
     * Annotations obtained from a MemberSelectTree are added to the parameter p.
     */
    private class JavariTreePreAnnotator extends SimpleTreeVisitor<Void, AnnotatedTypeMirror> {

        /**
         * Selects the appropriate annotation from the identifier, or
         * inherits it from the expression, as follows:
         *
         * <ul>
         *
         *  <li> If the identifier qualified type is annotated with
         *  {@code \@ReadOnly} or {@code \@Mutable}, the parameter
         *  receives the same annotation.
         *
         *  <li> If not, and if the expression qualified type has any
         *  annotation, the parameter receives those annotations.
         *
         *  <li> If the expression qualified type has no annotation,
         *  then the parameter receives a {@code @ThisMutable}
         *  annotation.
         *
         * </ul>
         *
         * @param node MemberSelectTree to be analyzed, of the form {@code expression . identifier}
         * @param p AnnotatedTypeMirror parameter to be annotated
         */
        @Override
        public Void visitMemberSelect(MemberSelectTree node, AnnotatedTypeMirror p) {

            AnnotatedTypeMirror exType = getAnnotatedType(node.getExpression());
            AnnotatedTypeMirror idType = fromElement(TreeUtils.elementFromUse(node));

            p.removeAnnotations(p.getAnnotations());
            if (idType.hasAnnotation(READONLY))
                p.addAnnotation(READONLY);
            else if (idType.hasAnnotation(MUTABLE))
                p.addAnnotation(MUTABLE);
            else if (hasImmutabilityAnnotation(exType))
                p.addAnnotation(getImmutabilityAnnotation(exType));
            else
                p.addAnnotation(THISMUTABLE);

            return super.visitMemberSelect(node, p);
        }

        /**
         * Returns list of annotations on the new class constructor;
         * resolves {code \@PolyRead} if applicable.
         */
        @Override
        public Void visitNewClass(NewClassTree node, AnnotatedTypeMirror p) {

            Element constructorElt = ((JCTree.JCNewClass)node).constructor;
            AnnotatedExecutableType executableType =
                (AnnotatedExecutableType) fromElement(constructorElt);
            AnnotatedTypeMirror receiverType = executableType.getReceiverType();
            
            if (receiverType.hasAnnotation(POLYREAD)) {
	            List<AnnotatedTypeMirror> parameterTypes = executableType.getParameterTypes();
	            List<AnnotatedTypeMirror> argumentTypes = atypes.getAnnotatedTypes(node.getArguments());
	
	            boolean allMutable = true, allPolyRead = true, allThisMutable = true;
	
	            for (int i = 0; i < parameterTypes.size(); i++) {
	                AnnotatedTypeMirror pType = parameterTypes.get(i);
	                if (pType.hasAnnotation(POLYREAD)) {
	                    AnnotatedTypeMirror aType = argumentTypes.get(i);
	
	                    if (aType.hasAnnotation(THISMUTABLE) || aType.hasAnnotation(POLYREAD))
	                        allMutable = false;
	
	                    if (aType.hasAnnotation(READONLY) || aType.hasAnnotation(QREADONLY)) {
	                        allMutable = false; allThisMutable = false;
	                    }
	
	                    if (!(aType.hasAnnotation(POLYREAD)
	                          && !aType.hasAnnotation(READONLY)
	                          && !aType.hasAnnotation(THISMUTABLE)
	                          && !aType.hasAnnotation(QREADONLY)))
	                        allPolyRead = false;
	                }
	            }
	            
	            if (!p.hasAnnotation(READONLY)) {
	            	if (allMutable) p.addAnnotation(MUTABLE);
	            	else if (allPolyRead) p.addAnnotation(POLYREAD);
	            	else if (allThisMutable) p.addAnnotation(THISMUTABLE);
	            	else p.addAnnotation(READONLY);
	            }
	            
            } else if (receiverType.hasAnnotation(READONLY))
            	p.addAnnotation(READONLY);
            else if (receiverType.hasAnnotation(MUTABLE) && !p.hasAnnotation(READONLY))
            	p.addAnnotation(MUTABLE);
               
            return super.visitNewClass(node, p);
        }
    }

    /**
     * Scanner responsible for adding implicit annotations to qualified types.
     *
     * <ul>
     *
     *   <li> Annotates unannotated receivers of qualified executable
     *   types with the qualified type owner's annotation.
     *
     *   <li> Annotates unannotated qualified declared types with
     *   their underlying type's element annotations.
     *
     *   <li> Annotates unnanotated qualified types whose elements
     *   correspond to fields with {@code \@ReadOnly}, if the enclosing
     *   element is readonly, or {@code @ThisMutable} otherwise.
     *
     * </ul>
     */
    private class JavariTypePostAnnotator extends AnnotatedTypeScanner<Void, Void> {

        /**
         * Annotates scanned qualified types that correspond to fields
         * and have no immutability annotations.
         *
         * If the enclosing qualified type is annotated with {@code
         * \@ReadOnly}, the scanned qualified type is annotated with {@code
         * \@ReadOnly}. Otherwise, the scanned qualified type is
         * annotated with {@code @ThisMutable}.
         */
        @Override
        public Void scan(AnnotatedTypeMirror type, Void p) {
            if (type != null && type.getElement() != null && !hasImmutabilityAnnotation(type)
                && type.getElement().getKind().isField()) {
                if (fromElement(type.getElement().getEnclosingElement()).hasAnnotation(READONLY))
                    type.addAnnotation(READONLY);
                else
                    type.addAnnotation(THISMUTABLE);
            }
            return super.scan(type, p);
        }

        /**
         * If the receiver has no annotation, annotates it with the
         * same annotation as the executable type owner.
         */
        @Override
        public Void visitExecutable(AnnotatedExecutableType type, Void p) {
            AnnotatedDeclaredType receiver = type.getReceiverType();
            if (!hasImmutabilityAnnotation(receiver)) {
                AnnotatedDeclaredType owner = (AnnotatedDeclaredType)
                    getAnnotatedType(type.getElement().getEnclosingElement());
                assert hasImmutabilityAnnotation(owner);
                receiver.addAnnotation(getImmutabilityAnnotation(owner));
            }
            return super.visitExecutable(type, p);
        }

        /**
         * If the declared type has no annotation, annotates it with
         * the same annotation as its underlying type's element.
         */
        @Override
        public Void visitDeclared(AnnotatedDeclaredType type, Void p) {
            if (!hasImmutabilityAnnotation(type)) {
                TypeElement tElt = (TypeElement) type.getUnderlyingType().asElement();
                AnnotatedTypeMirror tType = getAnnotatedType(tElt);
                assert hasImmutabilityAnnotation(tType);
                type.addAnnotation(getImmutabilityAnnotation(tType));
            }
            return super.visitDeclared(type, p);
        }
    }
}
