package checkers.types;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Set;

import javax.lang.model.element.AnnotationMirror;
import javax.lang.model.element.Name;

import checkers.util.AnnotationUtils;

/**
 * Represents a type qualifiers hierarchy.
 *
 * All method parameter annotations need to be type qualifiers recognized
 * within this hierarchy.
 */
public abstract class QualifierHierarchy {

    // **********************************************************************
    // Getter methods about this hierarchy
    // **********************************************************************

    /**
     * @return  the root (ultimate super) type qualifier in the hierarchy
     */
    public abstract AnnotationMirror getRootAnnotation();

    /**
     * Returns the names of all type qualifiers in this type qualifier
     * hierarchy
     *
     * @return the fully qualified name represented in this hierarchy
     */
    public abstract Set<Name> getTypeQualifiers();

    // **********************************************************************
    // Qualifier Hierarchy Queries
    // **********************************************************************

    /**
     * Tests whether anno1 is a super qualifier anno2, according to the type
     * qualifiers hierarchy.
     * 
     * @return true iff anno1 is a sub qualifier of anno2
     */
    public abstract boolean isSubtype(AnnotationMirror anno1, AnnotationMirror anno2);

    /**
     * Tests whether there is any annotation in lhs that is a super qualifier
     * of some annotation in rhs.
     *
     * @return true iff an annotation in lhs is a super of one in rhs
     **/
    // TODO: Explain this choice
    public boolean isSubtype(Collection<AnnotationMirror> rhs, Collection<AnnotationMirror> lhs) {
        for (AnnotationMirror lhsAnno : lhs)
            for (AnnotationMirror rhsAnno : rhs)
                if (isSubtype(rhsAnno, lhsAnno))
                    return true;
        return false;
    }

    /**
     * Check if the two annotations are in conflict.
     *
     * Two qualifiers are in conflict if both cannot annotate a type together.
     *
     * @return true iff anno1 and anno2 are in conflict
     */
    public abstract boolean inConflict(AnnotationMirror anno1, AnnotationMirror anno2);

    // **********************************************************************
    // Helper methods to extract annotations
    // **********************************************************************

    /**
     * Returns a subset of the input that contains any qualifiers in this
     * hierarchy
     * 
     * @return  the qualifiers in annos within this hierarchy
     */
    public List<AnnotationMirror> validQualifiers(Collection<AnnotationMirror> annos) {
        List<AnnotationMirror> results = new ArrayList<AnnotationMirror>();
        Set<Name> typeQualifiers = getTypeQualifiers();
        for (AnnotationMirror anno : annos) {
            if (typeQualifiers.contains(AnnotationUtils.annotationName(anno)))
                results.add(anno);
        }
        return results;
    }

    /**
     * Finds the first type qualifiers in this hierarchy in the given list of
     * qualifiers.
     *
     * @return  the qualifiers in annos in this hierarchy
     */
    public AnnotationMirror validQualifier(Collection<AnnotationMirror> annos) {
        List<AnnotationMirror> validQualifiers = validQualifiers(annos);
        return validQualifiers.isEmpty() ? null : validQualifiers.get(0);
    }
}
