package checkers.util;

import java.lang.annotation.Annotation;
import java.util.*;

import javax.annotation.processing.*;
import javax.lang.model.element.*;

import checkers.basetype.BaseTypeChecker;
import checkers.flow.Flow;
import checkers.metaquals.*;
import checkers.types.*;

import com.sun.source.tree.*;

/**
 * A checker for type qualifier systems that only checks subtyping
 * relationships.
 *
 * <p>
 *
 * The annotation(s) are specified on the command line, using an annotation
 * processor argument:
 *
 * <ul>
 * <li>{@code -Aquals}: specifies the annotations in the qualifier hierarchy
 * (as a comma-separated list of fully-qualified annotation names with no
 * spaces in between).  Only the annotation for one qualified subtype
 * hierarchy can be passed.</li>
 * </ul>
 *
 */
@SupportedOptions( { "quals" })
public class BasicChecker extends BaseTypeChecker {

    /** Whether flow is enabled if no lint option is provided. */
    protected static boolean FLOW_BY_DEFAULT = true;

    /** Qualifiers to infer. */
    private Set<AnnotationMirror> flowQuals = null;

    @Override
    public synchronized void init(ProcessingEnvironment p) {
        super.init(p);
        flowQuals = createFlowQualifiers(p);
    }

    private Set<AnnotationMirror> createFlowQualifiers(ProcessingEnvironment p) {
        AnnotationFactory annoFactory = new AnnotationFactory(p);

        Set<AnnotationMirror> flowQuals = new HashSet<AnnotationMirror>();
        for (Class<? extends Annotation> cl : getSupportedTypeQualifiers()) {
            flowQuals.add(annoFactory.fromClass(cl));
        }
        return flowQuals;
    }

    @Override
    public AnnotatedTypeFactory createFactory(CompilationUnitTree root) {
        return new UserTypeFactory(this, root);
    }

    @Override @SuppressWarnings("unchecked")
    protected Set<Class<? extends Annotation>> createSupportedTypeQualifiers() {

        // Is annotated?
        Set<Class<? extends Annotation>> supported = super.createSupportedTypeQualifiers();
        if (!supported.isEmpty()) {
            return supported;
        }

        String qualNames = env.getOptions().get("quals");
        if (qualNames == null)
            throw new Error("missing required option: -Aquals");

        Set<Class<? extends Annotation>> qualSet =
            new HashSet<Class<? extends Annotation>>();
        try {
            for (String qualName : qualNames.split(",")) {
                final Class<? extends Annotation> q =
                    (Class<? extends Annotation>)Class.forName(qualName);
                qualSet.add(q);
            }
        } catch (ClassNotFoundException e) {
            throw new Error(e);
        }
        return Collections.unmodifiableSet(qualSet);
    }

    /**
     * Extends {@link AnnotatedTypeFactory} to optionally use flow-sensitive
     * qualifier inference, qualifier polymorphism, and implicit annotations
     * via {@link ImplicitFor}.
     *
     * @see Flow
     */
    class UserTypeFactory extends AnnotatedTypeFactory {

        private final Flow flow;
        private final boolean useFlow;
        private final QualifierPolymorphism poly;
        private final TypeAnnotator typeAnnotator;
        private final TreeAnnotator treeAnnotator;

        /**
         * Creates a type factory for checking the given compilation unit with
         * respect to the given annotation.
         *
         * @param checker the checker to which this type factory belongs
         * @param root the compilation unit to scan
         */
        UserTypeFactory(BasicChecker checker, CompilationUnitTree root) {
            super(checker, root);

            treeAnnotator = new TreeAnnotator(checker);
            typeAnnotator = new TypeAnnotator(checker);

            poly = new QualifierPolymorphism(checker, this);

            flow = new Flow(checker, root, flowQuals, this);
            useFlow = flowQuals != null
                    && checker.getLintOption("flow", FLOW_BY_DEFAULT);

            // Apply flow-sensitive qualifier inference.
            if (useFlow) flow.scan(root, null);
        }

        @Override
        protected void annotateImplicit(Tree tree, AnnotatedTypeMirror type) {
            treeAnnotator.visit(tree, type);
            if (useFlow) {
                final AnnotationMirror inferred = flow.test(tree);
                if (inferred != null) {
                    type.clearAnnotations();
                    type.addAnnotation(inferred);
                }
            }
            poly.annotate(tree, type);
            typeAnnotator.visit(type);
        }

        @Override
        protected void annotateImplicit(Element elt, AnnotatedTypeMirror type) {
            poly.annotate(elt, type);
            typeAnnotator.visit(type);
        }
    }
}
